//    WeIRCd IRC server software - sub file
//    Copyright (C) 2007, 2008 Jonas Thiem
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.

#define USERQUIT_SIGPIPE 1
#define USERQUIT_QUIT 2
#define USERQUIT_EOF 3
#define USERQUIT_TIMEOUT 4
#define USERQUIT_ERROR 5
#define USERQUIT_SENDQEXCEEDED 6
#define USERQUIT_EXCESSFLOOD 7
#define USERQUIT_REGISTERTIMEOUT 8
#define USER_MAXSENDBUFFERS 20
#define USER_RECVBUFFERSIZE 1024
struct weircd_user_floodprotect {
	int floodpoints;
};
struct weircd_user_chaninfo {
	int chanid;
	bool iamop;
	bool iamhalfop;
	bool iamvoice;
};
struct weircd_user {
	unsigned char ip[17];
	unsigned char* sendbuffer[USER_MAXSENDBUFFERS];
	unsigned char recvbuffer[USER_RECVBUFFERSIZE];
	unsigned char host[COMPILEOPTION_MAXHOSTLEN];
	unsigned char pass[200];
	unsigned char* quitreason;
	long CONNECTtime;
	bool authed;//001 message passed
	long authpongvalue;
	int pleaseremove;
	bool ircop;
	long pleaseremove_time;
	bool sendbeforeremove;
	long lastnickchange;
	long logintime;
#ifdef SERVICES_ENABLED
	unsigned char* accountname;
#endif
	int socket;
	long ping_lastping;
	long ping_waitingforpong;
	long ping_pingval;
	struct weircd_user_floodprotect floodprotect;
	struct weircd_user_chaninfo* chan[COMPILEOPTION_MAXCHANSPERUSER];
	unsigned char nickname[COMPILEOPTION_MAXNICKLEN + 1];
	unsigned char suserid[COMPILEOPTION_MAXUSERIDLEN + 1];
	unsigned char realname[COMPILEOPTION_MAXREALNAMELEN + 1];
};

struct weircd_user* userpointers[COMPILEOPTION_MAXUSERS];
bool useraffected[COMPILEOPTION_MAXUSERS]; //needed for e.g. /who or tellaffectedusers();
int userslotcount = 0;

void weircdUserInit() {
	int r = 0;
	while (r < userslotcount) {
		userpointers[r] = NULL;
		r += 1;
	}
}
int userAdd(int socket) {
	int r = 0;
	while (r < COMPILEOPTION_MAXUSERS) {
		if (userpointers[r] == NULL) {
			userpointers[r] = malloc(sizeof(struct weircd_user));
			memset(userpointers[r], 0, sizeof(struct weircd_user));
			userpointers[r]->socket = socket;
			userpointers[r]->authed = false;
			userpointers[r]->sendbeforeremove = false;
			userpointers[r]->ircop = false;
			userpointers[r]->quitreason = NULL;
			if (r >= userslotcount) {userslotcount = r+1;}
			return r;
		}
		r += 1;
	}
	close(socket);//adding failed!
	return -1;
}
unsigned char* un(int uid) {
	if (userpointers[uid] != NULL) {
		return userpointers[uid]->nickname;
	}else{
		return NULL;
	}
}
int ubn(const unsigned char* nickname) {//UserByName()
	int r = 0;
	while (r < userslotcount) {
		if (userpointers[r] != NULL) {
			if (strcasecmp(userpointers[r]->nickname,nickname) == 0 && userpointers[r]->pleaseremove == false) {
				return r;
			}
		}
		r += 1;
	}
	return -1;
}
int uSocket(int uid) {
	return userpointers[uid]->socket;
}
bool isuserinchan(int userid, int vchanid) {
	if (userpointers[userid] == NULL) {return false;}
	if (userpointers[userid]->pleaseremove == true) {return false;}
	if (userpointers[userid]->authed == false) {return false;}
	int r = 0;
	while (r < COMPILEOPTION_MAXCHANSPERUSER) {
		if (userpointers[userid]->chan[r] != NULL) {
			if (userpointers[userid]->chan[r]->chanid == vchanid) {
				return true;
			}
		}
		r += 1;
	}
	return false;
}
int chanusercount(int vchanid) {
	int r = 0;int k = 0;
	while (r < COMPILEOPTION_MAXUSERS) {
		if (isuserinchan(r,vchanid) == true) {k += 1;}
		r += 1;
	}
	return k;
}
void removeemptychans() {
	int r = 0;
	while (r < COMPILEOPTION_MAXCHANS) {
		if (chans[r] != NULL) {
			if (chanusercount(r) <= 0) {
				free(chans[r]);
				chans[r] = NULL;
			}
		}
		r += 1;
	}
}
void tellaffectedusers(const unsigned char* stuff, int senderid) {
	int k = 0;
	int r;
	while (k < userslotcount) {
		useraffected[k] = false;
		k += 1;
	}
	k=0;
	while (k < COMPILEOPTION_MAXCHANSPERUSER) {
		if (userpointers[senderid]->chan[k] != NULL) {
			r = 0;
			while (r < userslotcount) {
				if (userpointers[r] != NULL && r != senderid) {
					if (userpointers[r]->pleaseremove == false) {
						if (isuserinchan(r,userpointers[senderid]->chan[k]->chanid) == true) {
							useraffected[r] = true;
						}
					}
				}
				r += 1;
			}
		}
		k += 1;
	}
	k=0;
	while (k < userslotcount) {
		if (useraffected[k] == true) {
			vsnd(stuff,k);
		}
		k += 1;
	}
}
void userQuit(int uid, int reason) {
	//if (userpointers[uid] == NULL) {printf("ERROR!!!!!!!!!!!(%d)\n",uid);return;}
	unsigned char response[501];
	if (userpointers[uid]->pleaseremove == false) {
		snprintf(&response[0],499,":%s!%s@%s QUIT :",userpointers[uid]->nickname,userpointers[uid]->suserid,userpointers[uid]->host);
		response[499] = 0;
		if (reason == USERQUIT_QUIT) {
			strncat(response,"Quit: ",499);
			response[499] = 0;
			if (userpointers[uid]->quitreason != NULL) {
				strncat(response,userpointers[uid]->quitreason,499);
				response[499] = 0;
				free(userpointers[uid]->quitreason);
				userpointers[uid]->quitreason = NULL;
			}
		}
		if (reason == USERQUIT_TIMEOUT) {
			strncat(response,"Ping timeout",499);
			response[499] = 0;
		}
		if (reason == USERQUIT_EOF) {
			strncat(response,"Client quit",499);
			response[499] = 0;
		}
		if (reason == USERQUIT_EXCESSFLOOD) {
			strncat(response,"Excess flood",499);
			response[499] = 0;
		}
		if (reason == USERQUIT_SENDQEXCEEDED) {
			strncat(response,"SendQ exceeded",499);
			response[499] = 0;
		}
		if (reason == USERQUIT_ERROR) {
			if (userpointers[uid]->quitreason != NULL) {
				strncat(response,userpointers[uid]->quitreason,499);
				response[499] = 0;
				free(userpointers[uid]->quitreason);
				userpointers[uid]->quitreason = NULL;
			}
		}
		tellaffectedusers(response,uid);
	}
	if (reason != USERQUIT_ERROR && reason != USERQUIT_TIMEOUT && reason != USERQUIT_SENDQEXCEEDED && reason != USERQUIT_EXCESSFLOOD && reason != USERQUIT_REGISTERTIMEOUT) {
		userpointers[uid]->pleaseremove = true;
		userpointers[uid]->sendbeforeremove = false;
		removeemptychans();
	}else{
		if (userpointers[uid]->pleaseremove == true) {return;}
		userpointers[uid]->pleaseremove = true;
		userpointers[uid]->pleaseremove_time = time(NULL);
		userpointers[uid]->sendbeforeremove = true;
		removeemptychans();
	}
}
void clearusers() {
	unsigned char snd[300];
	int r = 0;
	while (r < userslotcount) {
		if (userpointers[r] != NULL) {
			if (userpointers[r]->pleaseremove == true) {
				if (userpointers[r]->sendbeforeremove == false || userpointers[r]->pleaseremove_time + 2 < time(NULL) || userpointers[r]->sendbuffer[0] == NULL) {
					close(userpointers[r]->socket);
					if (userpointers[r]->quitreason != NULL) {
						free(userpointers[r]->quitreason);
						userpointers[r]->quitreason = NULL;
					}
					free(userpointers[r]);
					userpointers[r] = NULL;
				}
			}else{
				if (userpointers[r]->authed == true) {
					if (userpointers[r]->ping_waitingforpong <= 0) {
						if (userpointers[r]->ping_lastping + COMPILEOPTION_PINGFREQUENCY < time(NULL)) {
							userpointers[r]->ping_pingval = time(NULL);
							sprintf(&snd[0],"PING :%ld",userpointers[r]->ping_pingval);
							vsnd(snd,r);
							userpointers[r]->ping_waitingforpong = time(NULL);
						}
					}else{
						if (userpointers[r]->ping_waitingforpong + 20 < time(NULL)) {
							vsnd("ERROR :Closing link: ping timeout",r);
							userQuit(r,USERQUIT_TIMEOUT);
						}
					}
				}else{
					if (userpointers[r]->CONNECTtime + 10 < time(NULL)) {
						vsnd("ERROR :Closing link: Registration timeout",r);
						userQuit(r,USERQUIT_REGISTERTIMEOUT);
					}
				}
			}
		}
		r += 1;
	}
}
void tellallusersinchan(const unsigned char* stuff,int chanid) {
	int r = 0;
	while (r < userslotcount) {
		if (isuserinchan(r,chanid) == true) {
			vsnd(stuff,r);
		}
		r += 1;
	}
}
void tellallusersinchan_exclude(const unsigned char* stuff,int chanid,int excludeid) {
	int r = 0;
	while (r < userslotcount) {
		if (isuserinchan(r,chanid) == true && r != excludeid) {
			vsnd(stuff,r);
		}
		r += 1;
	}
}
unsigned char userprefix(int userid, int vchanid) {
	if (userpointers[userid] == NULL) {return 0;}
	if (userpointers[userid]->pleaseremove == true) {return 0;}
	if (userpointers[userid]->authed == false) {return 0;}
	if (isuserinchan(userid,vchanid) == false) {return 0;}
	int r = 0;
	while (r < COMPILEOPTION_MAXCHANSPERUSER) {
		if (userpointers[userid]->chan[r] != NULL) {
			if (userpointers[userid]->chan[r]->chanid == vchanid) {
				struct weircd_user_chaninfo* ch = userpointers[userid]->chan[r];
				if (ch->iamop == true) {
					return '@';
				}else{
					if (ch->iamhalfop == true) {
						return '%';
					}else{
						if (ch->iamvoice == true) {
							return '+';
						}else{
							return 0;
						}
					}
				}
			}
		}
		r += 1;
	}
	return 0;
}
void IOnotice(const unsigned char* str) {
	unsigned char stuff[500];
	int r = 0;
	while (r < userslotcount) {
		if (userpointers[r] != NULL) {
			if (userpointers[r]->pleaseremove == false && userpointers[r]->ircop == true) {
				snprintf(&stuff[0],499,":%s NOTICE %s :%s",cuo.svd,un(r),str);
				stuff[499] = 0;
				vsnd(stuff,r);
			}
		}
		r += 1;
	}
}
